from decimal import Decimal

from ..base import Field

class DecimalField(Field):

    def parse(self, value):
        if isinstance(value, Decimal):
            return value

        try:
            return Decimal(str(value))
        except Exception:
            self._raise_value_error(value)


from .exceptions import DefaultNotProvided, UnassignedOptionalFieldRequested, InvalidFieldArguments
from .serializer import DefaultSerializer

NOT_PROVIDED = object()


class Field:

    def __init__(self, default=NOT_PROVIDED, optional=False):

        if default is not NOT_PROVIDED and optional:
            raise InvalidFieldArguments(f"'{self.__class__.__name__}' cannot be both optional and have a default")

        self.default = default
        self.optional = optional
        self.field_name = 'Field'

    def get_default(self):

        if self.default is not NOT_PROVIDED:
            return self.default

        if self.optional:
            return NOT_PROVIDED

        raise DefaultNotProvided(f"Field '{self.field_name}' has no default specified and is not optional")

    def parse(self, value):
        return value

    def default_serializer(self, value):
        return str(value)

    @classmethod
    def is_field(cls, value):
        return isinstance(value, cls)

    def _raise_value_error(self, value):
        raise ValueError(
            f'Field "{self.field_name}" with value "{value}" could not be parsed as a {self.__class__.__name__}') from None


class FieldValue:

    def __init__(self, field: Field, value=NOT_PROVIDED):
        self.field = field
        self.value = value

    def get(self):
        if self.value is NOT_PROVIDED:
            self.value = self.field.get_default()

        if self.is_not_provided():
            raise UnassignedOptionalFieldRequested(
                f"Optional Field '{self.field.field_name}' was not assigned and was requested")

        return self.value

    def set(self, value):
        if value is not NOT_PROVIDED:
            self.value = self.field.parse(value)

    def is_not_provided(self):
        return self.value is NOT_PROVIDED


class ModelMeta(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        module = attrs.pop('__module__')
        base_attrs = {'__module__': module}

        fields = {}

        for base in bases:
            base_model_meta = getattr(base, '_model_meta', None)
            if base_model_meta:
                fields = {**fields, **base_model_meta.get('fields', {})}


        for key, value in list(attrs.items()):
            if Field.is_field(value):
                fields[key] = value
                fields[key].field_name = key
            else:
                base_attrs[key] = value

        new_class = super().__new__(cls, name, bases, base_attrs, **kwargs)  # noqa

        new_class._model_meta = {
            'fields': fields,
        }
        return new_class


class Model(metaclass=ModelMeta):

    def __init__(self, dict_=None, **kwargs):
        if dict_ is None:
            dict_ = {}
        source = {**dict_, **kwargs}

        self._field_values = {}

        for field_name in self._model_meta['fields']:
            field: Field = self._model_meta['fields'][field_name]
            field_value = FieldValue(field=field)
            try:
                value = source.pop(field_name)
            except KeyError:
                value = field.get_default()

            field_value.set(value)

            self._field_values[field_name] = field_value

    def __getattribute__(self, item):
        if item in ['_model_meta', '_field_values']:
            return super().__getattribute__(item)

        if item not in self._model_meta['fields']:
            return super().__getattribute__(item)

        field_value: FieldValue = self._field_values[item]
        return field_value.get()

    def __setattr__(self, key, value):

        if key not in self._model_meta['fields']:
            super().__setattr__(key, value)
            return

        field_value: FieldValue = self._field_values[key]
        field_value.set(value)

    def serialize(self, serializer=DefaultSerializer):
        return serializer.serialize(self)



from ..base import Field

class IntegerField(Field):

    def parse(self, value):
        try:
            return int(value)
        except (TypeError, ValueError):
            self._raise_value_error(value)

    def default_serializer(self, value):
        return value


from enum import Enum

from ..exceptions import InvalidFieldArguments
from ..base import Field, NOT_PROVIDED


class EnumField(Field):

    def __init__(self, enum_type=NOT_PROVIDED, default=NOT_PROVIDED, optional=False):
        super().__init__(default, optional)

        if enum_type is NOT_PROVIDED:
            raise InvalidFieldArguments('An EnumField requires a enum_type to be specified')

        if not issubclass(enum_type, Enum):
            raise TypeError(
                f'The enum_type specified must inherit from Enum. You have provided enum_type of type {type(enum_type)}')

        self.enum_type = enum_type

    def parse(self, value):
        try:
            return self.enum_type(value)  # noqa
        except ValueError as e:
            raise ValueError(
                f'Field "{self.field_name}" with value "{value}" could not be parsed as a {self.enum_type.__name__}') from None

    def default_serializer(self, value):
        return value.value



from .boolean import BooleanField
from .string import StringField
from .integer import IntegerField
from .float import FloatField
from .decimal import DecimalField
from .datetime import DateTimeField
from .enum import EnumField
from .list import ListField
from .model import ModelField


from ..exceptions import InvalidFieldArguments
from ..base import Field, Model, NOT_PROVIDED


class ModelField(Field):

    def __init__(self, model_class=NOT_PROVIDED, default=NOT_PROVIDED, optional=False):
        super().__init__(default, optional)

        if model_class is NOT_PROVIDED:
            raise InvalidFieldArguments('A ModelField requires a model_class to be specified')

        if not issubclass(model_class, Model):
            raise TypeError(
                f'The ModelClass specified must inherit from Model. You have provided model_class of type {type(model_class)}')

        self.ModelClass = model_class

    def parse(self, value):
        if isinstance(value, dict):
            return self.ModelClass(value)

        if not isinstance(value, self.ModelClass):
            self._raise_value_error(value)

        return value


    def serialize(self, value, serializer=NOT_PROVIDED):
        if serializer is NOT_PROVIDED:
            raise RuntimeError("A serializer is expected to serialize a Model")

        return serializer.serialize(value)


from .exceptions import UnassignedOptionalFieldRequested


class DefaultSerializer:

    @staticmethod
    def serialize(model):

        output = {}

        for field_name, field_value in model._field_values.items():
            try:
                output[field_name] = DefaultSerializer.serialize_field(field_value)
            except UnassignedOptionalFieldRequested:
                pass

        return output

    @staticmethod
    def serialize_field(field_value):
        custom_serializer = getattr(field_value.field, 'serialize', None)

        if custom_serializer:
            return custom_serializer(field_value.get(), serializer=DefaultSerializer)

        return field_value.field.default_serializer(field_value.get())



from ..base import Field

class BooleanField(Field):

    def parse(self, value):
        if isinstance(value, bool):
            return value

        if isinstance(value, str):
            if value.lower() in ['true', 'True']:
                return True
            elif value.lower() in ['false', 'False']:
                return False

        self._raise_value_error(value)

    def default_serializer(self, value):
        return value


from collections.abc import MutableSequence

from ..exceptions import InvalidFieldArguments
from ..base import Field, NOT_PROVIDED
from .model import ModelField


class ListField(Field):

    def __init__(self, list_type=NOT_PROVIDED, default=NOT_PROVIDED, optional=False):
        if list_type is NOT_PROVIDED:
            raise InvalidFieldArguments('An ListField requires a list_type to be specified')

        if not isinstance(list_type, Field):
            raise TypeError(
                f'The list_type specified must be an instance of a subclass of Field. You have provided list_type of type {type(list_type)}')

        self.ListType = list_type

        default_value = default if default is not NOT_PROVIDED else self.parse([])

        super().__init__(default_value, optional)

    def parse(self, value):
        if not isinstance(value, list) and not isinstance(value, TypedFieldList):
            raise TypeError(
                f'Field "{self.field_name}" expects a value of type list, however {type(value)} was provided')

        typed_field_list = TypedFieldList(self.ListType)
        typed_field_list.extend(value)
        return typed_field_list

    def serialize(self, value, serializer=NOT_PROVIDED):
        if serializer is NOT_PROVIDED:
            return [self.ListType.default_serializer(v) for v in value]

        if isinstance(self.ListType, ModelField):
            return [serializer.serialize(v) for v in value]

        return [serializer.serialize_field(v) for v in value]


class TypedFieldList(MutableSequence):

    def __init__(self, field_type, *args):
        self.field_type: Field = field_type
        self.list = list()
        self.extend(list(args))

    def __len__(self):
        return len(self.list)

    def __getitem__(self, i):
        return self.list[i]

    def __delitem__(self, i):
        del self.list[i]

    def __setitem__(self, i, v):
        self.list[i] = self.field_type.parse(v)

    def insert(self, i, v):
        self.list.insert(i, self.field_type.parse(v))

    def __str__(self):
        return str(self.list)



import pendulum

from ..base import Field, NOT_PROVIDED


class DateTimeField(Field):

    AUTO_NOW = 'AUTO_NOW'

    def __init__(self, default=NOT_PROVIDED, optional=False, tz='UTC'):
        super().__init__(default, optional)
        self.tz = tz

    def parse(self, value):
        if isinstance(value, pendulum.DateTime):
            return self._set_tz(value)

        try:
            return self._set_tz(pendulum.parse(value))
        except (pendulum.exceptions.ParserError, TypeError):
            self._raise_value_error(value)

    def _set_tz(self, dt):
        return dt.in_tz(tz=self.tz)

    def get_default(self):
        if self.default == self.AUTO_NOW:
            return self._set_tz(pendulum.now())

        return self.parse(super().get_default())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typed_models', 'typed_models.fields']

package_data = \
{'': ['*']}

install_requires = \
['pendulum>=2.1.2,<3.0.0']

setup_kwargs = {
    'name': 'typed-models',
    'version': '0.2.0',
    'description': 'Strongly typed, lightweight, serializable models for Python',
    'long_description': "Contributing\n============\n\nContributions are welcome.\n\nGetting started\n---------------\n\nTo work on the typed-models codebase, you'll want to clone the project locally\nand install the required dependencies via `poetry <https://poetry.eustace.io>`_.\n\n.. code-block:: bash\n\n    $ git clone git@github.com:a2d24/typed-models.git\n    $ poetry install",
    'author': 'Imtiaz Mangerah',
    'author_email': 'Imtiaz_Mangerah@a2d24.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/a2d24/typed-models',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
